:- use_module(library(clpq)).
:- use_module(library(lists)).
:- use_module(qdecproc).


%% bin_clause(p(X1,...,Xn),C,p(Y1,...,Yn))
%% hyp : vars(C) in Xs+Ys

% find_filter(+BinClause,-Filter,-DNlog,-BDMGH)
% -- with the projection of the constraints on the head variables
find_filter_projection(BinClause,Filter,DNlog,BDMGH) :-
	BinClause=bin(H,C,B),
	functor(H,P,N), H=..[P|Xs],
	functor(B,P,N), B=..[P|Ys],
	qproject(Xs,C,Vars,Delta),
	list1N(N,L),
	findall(S,suset(S,L),Taus),
	find_filter_aux(Taus,Xs,Ys,C,Vars,Delta,Filter,DNlog,BDMGH).
% -- with true
find_filter_true(BinClause,Filter,DNlog,BDMGH) :-
	BinClause=bin(H,C,B),
	functor(H,P,N), H=..[P|Xs],
	functor(B,P,N), B=..[P|Ys],
	Delta=[],copy_term(Xs,Vars),
	list1N(N,L),
	findall(S,suset(S,L),Taus),
	Taus=[[]|Taus2],append(Taus2,[[]],Taus3),
	find_filter_aux(Taus3,Xs,Ys,C,Vars,Delta,Filter,DNlog,BDMGH).


list1N(N,L) :- list1N(1,N,L).
list1N(I,N,L) :- 
	(   I =< N 
	->  J is I+1,
	    L=[I|L1],
	    list1N(J,N,L1)
	;   L=[]).
      
suset([],_B).
suset([X|Xs],[X|Ys]) :- suset(Xs,Ys).
suset([X|Xs],[_|Zs]) :- suset([X|Xs],Zs).	

%find_filter_aux(+Taus,+Xs,+Ys,+C,+Vars,+Delta,-Filter,-DNlog,-BDMGH)
find_filter_aux([Tau|_Taus],Xs,Ys,C,Vars,Delta,Tau-(CalVars,Deltacv),DNlog,BDMGH) :-
	cal_args(Tau,Vars,CalVars),
	qproject(CalVars,Delta,CalVars,Deltacv),
	dnlog_formula(Xs,Ys,C,CalVars,Deltacv,Tau,DNlog),
	b_more_general_h(Xs,Ys,C,CalVars,Deltacv,Tau,BDMGH).
find_filter_aux([_|Taus],Xs,Ys,C,Vars,Delta,Filter,DNlog,BDMGH) :-
	find_filter_aux(Taus,Xs,Ys,C,Vars,Delta,Filter,DNlog,BDMGH).

b_more_general_h(Xs,Ys,C,Vars,Delta,Tau,(C1,BsD)) :-
    % Ys,C is Tau-Vars-Delta more general than Xs,C
    % ie  (1) Ys,C_{bar(Tau)} is more general than Xs,C_{bar(Tau)}
    cond1(Xs,Ys,C,Vars,Delta,Tau,C1),
    % and (2) Ys,C satisfies Tau-Vars-Delta
    b_satisfies_delta(Ys,C,Tau,Vars,Delta,BsD).

cond1(Xs2,Ys2,C2,_Vars,_Delta,Tau,Cond1) :-
    copy_term(Xs2-Ys2-C2,Xs-Ys-C),
    length(Xs,N), complement(Tau,N,Taubar),
    cal_args(Taubar,Xs,CalXs), qproject(CalXs,C,CalXs,Cxs), qdecproc:list_round(Cxs,Dxs),
    cal_args(Taubar,Ys,CalYs), qproject(CalYs,C,CalYs,Cys), qdecproc:list_round(Cys,Dys),
    CalXs=CalYs,
    univ_quantif(CalXs,(Dxs -> Dys),Cond1).

b_satisfies_delta(Ys,C,Tau,Vars,Delta,BsD) :-
    cal_args(Tau,Ys,CalYs),
    qproject(CalYs,C,Ys2,C2),
    qdecproc:list_round(C2,C2r),
    qdecproc:list_round(Delta,Deltar),
    Ys2=Vars,
    univ_quantif(Vars,(C2r -> Deltar),BsD).

% b_satisfies_delta([N1,T1,P1],[N=N1+1,T1=2*T,P1=P,T>=1,P>=2],[2,3],[B,C],[B>=1,C>=2],BsD).
% BsD = forall(B, forall(C, (ac(C>=2), ac(B>=2)->ac(B>=1), ac(C>=2))))

complement(Tau,N,Taubar) :-
    findall(I,(between(1,N,I),\+ member(I,Tau)),Taubar).

dnlog_formula(Xs,Ys,C,CalVars,Delta,Tau,DNlog) :-
	cal_args(Tau,Xs,CalXs),
	cal_args(Tau,Ys,CalYs),
	sat(CalXs,CalVars,Delta,NewCalXs,SatXs),
	NewCalXs=CalXs,
	sat(CalYs,CalVars,Delta,NewCalYs,SatYs),
	NewCalYs=CalYs,
	qdecproc:list_round(C,RoundC),
	%
	exist_quantif(NewCalYs,RoundC,ExistYRoundC),
	qdecproc:list_round(SatXs,RoundSatXs),
	MiddleLeftFormula=(RoundSatXs -> ExistYRoundC),
	univ_quantif(NewCalXs,MiddleLeftFormula,QuantifiedMiddleLeftFormula),
	Formula1=(RoundC -> QuantifiedMiddleLeftFormula),
	append(Xs,Ys,Zs),
	univ_quantif(Zs,Formula1,DNlog1),
	%
	qdecproc:list_round(SatYs,RoundSatYs),
	Formula2=(RoundC -> RoundSatYs),
	univ_quantif(Zs,Formula2,DNlog2),
	%
	DNlog=(DNlog1,DNlog2).
	

exist_quantif([],F,F).
exist_quantif([X|Xs],F,exists(X,G)) :- exist_quantif(Xs,F,G).

univ_quantif([],F,F).
univ_quantif([X|Xs],F,forall(X,G)) :- univ_quantif(Xs,F,G).
	
cal_args([],_Xs,[]).
cal_args([I|Is],Xs,[Xi|XIs]) :- fnth1(I,Xs,Xi),cal_args(Is,Xs,XIs).

fnth1(I,[_|Xs],X):- I>1,!,J is I-1,fnth1(J,Xs,X).
fnth1(1,[X|_],X).

%%%%
sat(CalXs,CalVars,Delta,NewCalXs,SatXs) :-
	copy_term(CalVars-Delta,CalVars2-Delta2),
	eqns(CalXs,CalVars2,Eqns),
	append(Eqns,Delta2,Cons),
	qproject(CalXs,Cons,NewCalXs,SatXs).

eqns([],[],[]).
eqns([X|Xs],[Y|Ys],[X=Y|Eqns]) :- eqns(Xs,Ys,Eqns).


%%%% 	instancier(T,Xs,C)
instanciate(T,Xs,C) :-
	cal_args(T,Xs,DNArgs),
	qproject(DNArgs,C,DNArgs,Cp),
	qdecproc:tell_cs(Cp),
	give_values(DNArgs).

give_values([]).
give_values([X|Xs]) :- 
	(  inf(X,X) -> 
	   true
	;   
	    (	sup(X,X) ->
		true
	    ;	
	        {X=0})),
	give_values(Xs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% some binary clp(Q) clauses
% NB : the variables in the constraint
% have to be included in the union of the variables
% of the head and the body

:- style_check(-singleton).

bin_clause(p(X),[],p(Y)).
bin_clause(p(X),[X=Y],p(Y)).
bin_clause(p(X),[X=0],p(Y)).
bin_clause(p(X),[X=0,Y=0],p(Y)).
bin_clause(p(X),[X=0,Y=1],p(Y)).
bin_clause(p(X),[X>=0,Y=1],p(Y)).
bin_clause(p(X),[X>=0,Y>=1],p(Y)).
bin_clause(p(X),[X>=0,Y>= -1],p(Y)).
bin_clause(p(X),[X>=1,Y=< 0],p(Y)).
bin_clause(p(X),[X=Y+1,Y>=0],p(Y)).
%%%bin_clause(p(X),[X=Y+1],p(Y)).
%%%bin_clause(p(X),[X>=2],p(Y)).
%%bin_clause(p(X),[X>=2,Y>=3],p(Y)).
%%%bin_clause(p(X),[X>=2,Y>=1],p(Y)).
%%%bin_clause(p(X),[X>=1,Y=< 0],p(Y)).
%%%bin_clause(p(X),[X>=0,Y=< 0],p(Y)).

bin_clause(p(X,Z),[X=Y+1,Y>=0],p(Y,T)).
bin_clause(p(X,Z),[X=Y+1,Y>=0,Z=T],p(Y,T)).
bin_clause(p(X,Z),[X=Y+1,Y>=0,Z+1=T],p(Y,T)).
bin_clause(p(X,Z),[X=Y+1,Y>=0,Z+1=T,T>=0],p(Y,T)).
bin_clause(p(X,Z),[X=Y+1,Y>=0,Z=T+1,T>=0],p(Y,T)).
bin_clause(p(X1,X2),[X1>=X2,Y1=X1+1,Y2=X2],p(Y1,Y2)).                         % ex 4.33
bin_clause(p(X1,X2),[X1=<X2,Y1=X1+1,Y2=X2],p(Y1,Y2)).                         % ex 4.34 
bin_clause(pow2(N,T,P),[N=N1+1,N1>=0,T1=2*T,P1=P,T>=1,P>=2],pow2(N1,T1,P1)).  % ex 4.35


go :-
	%%
	bin_clause(H,C,B),BinClause=(H :- C,B),
	nl,nl,write('-------------------------------------------'),nl,
	\+ \+ (numbervars(BinClause,0,_),print(BinClause),nl),
	find_filter_projection(bin(H,C,B),Filter,DNlog,BDMGH),
	%find_filter_true(bin(H,C,B),Filter,DNlog,BDMGH),
       	Filter=Tau-_,
	write('- checking the filter: DELTA ='),
	\+ \+ (numbervars(Filter,0,_), print(Filter),nl),
	%%
	write('the formulas DNlogs: '),nl,
	DNlog=(DNlog1,DNlog2),
	\+ \+ (numbervars(DNlog,0,_),
	       write('  DNlog1: '),print(DNlog1),nl,
	       write('  DNlog2: '),print(DNlog2),nl),
	(   test(DNlog1,'DNlog1')
	->  (   test(DNlog2,'DNlog2')
	    ->  write('the formulas BDMGH \'B delta more general than H\': '),nl,
		BDMGH=(BpMGHp,BsD),
		\+ \+ (numbervars(BDMGH,0,_),
		       write('  Br mg Hr: '),print(BpMGHp),nl,
		       write('  B sat Dt: '),print(BsD),nl),
		(   test(BpMGHp,'Br mg Hr')
		->  test(BsD,'B sat Dt'),
		    write('** filter OK: DNlogs and BDMGH are true'), nl,
		    H=..[P|Xs],qproject(Xs,C,Xs,Cproj),
		    write('** any query delta more general than '),
		    \+ \+ (numbervars(H-B,0,_),print(H-Cproj),
			   write(' does not terminate wrt '),print((H :- C, B)),nl), 
		    write('** in particular, '), 
		    copy_term(Xs-C,Xsc-Cc),instanciate(Tau,Xsc,Cc),H2=..[P|Xsc],
		    term_variables(Xsc,Vars),qproject(Vars,Cc,Vars,Ccp),
		    \+ \+ (numbervars(H2-Ccp,0,_),print(H2-Ccp),
			   write(' does not terminate wrt '),
			   numbervars(H-C-B,0,_),print((H :- C, B)))
		;   write('No looping query found'))
	    ;   write('No looping query found'))
	;   write('No looping query found')),
	nl,nl,
        fail.
go.

test(F,Name) :-
	write(Name),
	(   qdecproc(F,ac(0<1)) 
	->  write(' is true'),nl
	;   write(' is false'),nl,fail).

/* SP:
between(A,B,A) :- A =< B.
between(A,B,X) :- A < B, A1 is A+1, between(A1,B,X).
:- use_module(library(terms)).
*/

:- go.
